#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 6103 $
$LastChangedDate:: 2022-08-22 16:33:52 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import glob
import os
import json
import pandas as pd
from symspace.utils.decorators import try_catch
from symspace.osutils.pathutils import Paths
import symspace.osutils.processutils as pu
from componentspace.report.junit_reader import JUnitReader
import componentspace.osutils.fileutils as fu
from componentspace.osutils.paths import ComponentPaths
from componentspace.builders.scripts import write_external_version


class JSONReport:
    """Class to report the component versions"""

    def __init__(self, build_info: str):
        """Initialize json report class

        Parameters
        ----------
        build_info : str
            Path to the json file, to read and write the component versions
        """
        self.build_info_json = build_info
        if os.path.isfile(build_info):
            '''Check if build information is available'''
            self.build = json.loads(open(build_info).read())
            self.cleanUp()
        else:
            self.build = {}

    def getBuildTestComponents(self, full_build: bool) -> list:
        """
        Collects the all components for which a test should be built

        Parameters
        ----------
        full_build : bool
            Build all component tests if true

        Returns
        -------
        paths_list : list
            paths of components
        """
        if full_build:
            '''build tests for all components'''
            paths_list = self.build.keys()
            '''set all BuildTest flags to true'''
            for key in self.build.keys():
                self.addBuildTests(key, True)
        else:
            paths_list = []
            for key in self.build.keys():
                '''build tests only for components where the BuildTest flags equal True '''
                if self.getBuildTests(key):
                    paths_list.append(key)
        return paths_list

    def _add_item(self, comp_path: str, key: str, value):
        if comp_path not in self.build:
            self.build[comp_path] = dict()
        self.build[comp_path][key] = value

    def _get_item(self, comp_path: str, key: str):
        value = None
        if comp_path in self.build:
            if key in self.build[comp_path]:
                value = self.build[comp_path][key]
        return value

    def addRevision(self, comp_path: str, revision: int):
        """
        Add the component revision

        Parameters
        ----------
        comp_path : str
            Component key
        revision : int
            New component revision
        """
        self._add_item(comp_path, 'Revision', revision)

    def addBuildTests(self, comp_path: str, build_tests: bool = True):
        """
        Add the component revision

        Parameters
        ----------
        comp_path : str
            Component key
        build_tests : bool
            Build component test, False (Default)
        """
        self._add_item(comp_path, 'BuildTests', build_tests)

    def addPackages(self, comp_path: str, packages: dict = None):
        if packages is not None:
            self._add_item(comp_path, 'Packages', packages)
        else:
            pkgs_json = ComponentPaths.getPackagesFile(comp_path)
            if os.path.isfile(pkgs_json):
                packages = json.loads(open(pkgs_json).read())
                self._add_item(comp_path, 'Packages', packages)

    def addVersion(self, comp_path: str, version: str = None):
        if version is not None:
            self._add_item(comp_path, 'Version', version)
        else:
            version_file = ComponentPaths.getVersionFile(comp_path)
            if os.path.isfile(version_file):
                version = self.readVersion(comp_path)
                self._add_item(comp_path, 'Version', version)

    @staticmethod
    @try_catch
    def writeVersion(comp_path: str):
        mop_files = glob.glob(os.path.join(comp_path, '*.mop'))
        pu.csymspaceRun(write_external_version.__file__, cmd_args=mop_files)

    @staticmethod
    def readVersion(comp_path: str) -> str:
        version_file = ComponentPaths.getVersionFile(comp_path)
        if os.path.isfile(version_file):
            version = str(json.loads(open(version_file).read()))
        else:
            version = '0.0'
        return version

    def removeEntry(self, comp_path: str):
        """
        Remove the component entry

        Parameters
        ----------
        comp_path : str
            Component key
        """
        self.build.pop(comp_path, None)

    def getVersion(self, comp_path: str) -> str:
        """
        Get the component version

        Parameters
        ----------
        comp_path : str
            Component key

        Returns
        -------
        str
        """
        version = self._get_item(comp_path, 'Version')
        version = '0.0' if version is None else str(version)
        return version

    def getRevision(self, comp_path: str) -> int:
        """
        Get the component revision

        Parameters
        ----------
        comp_path : str
            Component key

        Returns
        -------
        int
        """
        revision = self._get_item(comp_path, 'Revision')
        revision = 0 if revision is None else revision
        return revision

    def getPackages(self, comp_path: str) -> dict:
        """
        Get the component packages and versions

        Parameters
        ----------
        comp_path : str
            Component key

        Returns
        -------
        packages : dict
            pkg_name, version
        """
        packages = self._get_item(comp_path, 'Packages')
        packages = dict() if packages is None else packages
        return packages

    def getBuildTests(self, comp_path: str) -> bool:
        """
        Get the build test flag

        Parameters
        ----------
        comp_path : str
            Component key

        Returns
        -------
        bool
        """
        build_tests = self._get_item(comp_path, 'BuildTests')
        build_tests = False if build_tests is None else build_tests
        return build_tests

    def writeJSON(self):
        """
        Write a json file from the collected component data
        """
        json_file = self.build_info_json

        with open(json_file, 'w') as fh:
            json.dump(self.build, fh, indent=4, sort_keys=True)

    def writeXLS(self, xls_path: str):
        """
        Write a excel file from the collected component data

        Parameters
        ----------
        xls_path : str
        """

        df = pd.DataFrame.from_dict(self.build).transpose()
        df.to_excel(xls_path)

    def appendXLS(self, xls_writer, sheet_name='Sheet1', key=None, value=None, comperator=None):
        if key is None:
            df = pd.DataFrame.from_dict(self.build).transpose()
            df.to_excel(xls_writer, sheet_name)

        else:
            print("comperator: ", comperator)
            if comperator is "<":
                subdictionary = {k: v for k, v in self.build.items() if key in v and float(v[key]) < float(value)}
            elif comperator is "=":
                print(key, value)
                subdictionary = {k: v for k, v in self.build.items() if key in v and v[key] == value}
            elif comperator is ">":
                subdictionary = {k: v for k, v in self.build.items() if key in v and float(v[key]) > float(value)}

            df = pd.DataFrame.from_dict(subdictionary).transpose()
            df.to_excel(xls_writer, sheet_name)

    def updateFromJUnit(self, junit_xml_file: str):
        """
        Update the build_info by searching for failed and succeeded tests of SymProjectTestCases

        Parameters
        ----------
        junit_xml_file: str
        """
        ok_list = []
        error_list = []

        if os.path.isfile(junit_xml_file):
            print('Analyzing TestCases from', junit_xml_file)
            reader = JUnitReader(junit_xml_file)
            test_ok, test_error = reader.findSyMProjectTestErrors()
            ok_list.extend(test_ok)
            error_list.extend(test_error)
            # os.remove(xml_file)

            for key in self.build.keys():
                '''loop over all registered components'''
                disp_path = key.replace(Paths.getComponentSpaceRoot(), '.')

                if key in ok_list:
                    '''don't force a new build of the test if the test was ok'''
                    self.addBuildTests(key, False)
                elif key in error_list:
                    '''force a new build of the test if the test failed'''
                    print('Schedule component and test rebuild: {}'.format(disp_path))
                    self.addBuildTests(key, True)
                    '''Something might be wrong with the component, 
                       hence schedule a new build of the component in the next run
                    '''
                    self.build[key]['Revision'] = 0
                else:
                    '''no test results for this component found, may happen in an incremental build'''
                    print('No unittest report found: {}'.format(disp_path))

        self.writeJSON()

    def cleanUp(self):
        """
        Perform a clean-up of the dataset

        Remove no more existing directories from the dataset

        """
        start_path = os.path.dirname(self.build_info_json)
        file_paths = fu.searchBuildComponentScripts(start_path)

        component_paths = [os.path.dirname(file) for file in file_paths]
        registered_paths = self.build.keys()

        remove_paths = list(set(registered_paths).difference(set(component_paths)))

        for comp_path in remove_paths:
            if comp_path in registered_paths:
                self.removeEntry(comp_path)
