#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 5027 $
$LastChangedDate:: 2021-10-12 15:12:32 +0200#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import optparse
import datetime
import glob
from os.path import join as pathjoin
from componentspace.report.junit import JUnitReport
from symspace.osutils.pathutils import Paths
from symspace.osutils.pathutils import CMDParser
from symspace.osutils.multiprocessing import ExternalMultiprocessing
from componentspace.testing.scripts import build_external_project_test
from componentspace.testing.abstract_test_builder import AbstractTestBuilder


class ProjectTestBuilder(AbstractTestBuilder):
    xls_file_start = 'test_'
    xls_file_ending = '.sym.xlsx'
    mop_file_ending = '_sym.mop'
    build_reports_dir = os.path.join(Paths.getCSTestReportsDir(), 'build_project_tests')
    run_reports_dir = os.path.join(Paths.getCSTestReportsDir(), 'run_project_tests')

    @classmethod
    def get_search_pattern(cls):
        search_pattern = '\\**\\_test_*{}'.format(cls.mop_file_ending)
        return search_pattern

    # Todo remove start_path from __init__
    def __init__(self, start_path: str, timeout: float = 300):
        """
        Class collects and builds SyMProject tests from excel files

        Parameters
        ----------
        start_path : str
            Path to the directory where build starts.
            Tests in all subdirectories will be built.
        timeout : float
            Timeout (in seconds) for single test build script
        """
        ''' unittest files'''
        build_results_xml = pathjoin(start_path, 'build_project_test_results-junit.xml')
        junit_build = JUnitReport('componentspace.create_tests.Build', build_results_xml, '')
        super().__init__(start_path, timeout, junit_build=junit_build)

    @classmethod
    def fromCMDParser(cls) -> 'ProjectTestBuilder':
        """
        Commandline parser to create instance of this class

        Commandline Arguments
        ---------------------

        --startDir : str
            The search for test files starts from this directory (Optional).
            Default is the ComponentSpace root directory.

        Returns
        -------
        builder : ComponentTestBuilder
        """

        parser = optparse.OptionParser("usage: %prog [options]")
        parser.add_option("-d", "--startDir", dest="start_dir", default=Paths.getComponentSpaceRoot(),
                          type="string", help="The search for components starts from this directory, "
                                              "default is the ComponentSpace root directory.")

        args = CMDParser.jepSysArgs()
        (options, args) = parser.parse_args(args)

        start_dir = CMDParser.jepArgPath(options.start_dir)

        print('Start Path: {}'.format(start_dir))

        builder = cls(start_dir)
        return builder

    def startBuild(self):
        """
        Try to build test projects for all test files
        """
        print('\nStarting full build of project tests...\n')

        print('Start Directory: {}'.format(self.start_path))

        file_list = glob.glob(self.start_path + '/**/{}*{}'.format(self.xls_file_start, self.xls_file_ending),
                              recursive=True)

        self.buildTests(file_list)

    def buildTests(self, file_list: list):
        """
        Try to build test projects for all test files in the files_list

        This method could be killed by an external process,
        hence the workaround make sure the intermediate process is stored on the file system.

        Parameters
        ----------
        file_list : list
            List of files to build tests for

        """
        n_tests = len(list(file_list))
        print('\nBuilding {} Excel tests'.format(n_tests))

        self.junit_build.startTimer()

        xmp = ExternalMultiprocessing(max_cpu_percent=75, max_ram_percent=75, timeout=self.timeout)

        disp_paths = []

        for index, xls_file in enumerate(file_list):

            disp_path = xls_file.replace(Paths.getComponentSpaceRoot(), '.')
            disp_paths.append(disp_path)

            prj_name = os.path.basename(xls_file).replace(self.xls_file_ending, '')
            test_dir = os.path.join(os.path.dirname(xls_file), prj_name)
            mop_file = os.path.join(test_dir, '_{}{}'.format(prj_name, self.mop_file_ending))
            build_tests_py = os.path.join(test_dir, 'build_tests.py')
            timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
            print('{} Build Test {}/{} ==> {}'.format(timestamp, index + 1, n_tests, disp_path))

            os.makedirs(os.path.join(test_dir, 'tests'), exist_ok=True)

            if os.path.isfile(build_tests_py):
                script_py = build_tests_py
            else:
                script_py = build_external_project_test.__file__

            cmd_args = [Paths.getCSyMSpaceExe(), script_py, mop_file, xls_file]
            xmp.startProcess(cmd_args, wait_to_start=False, wait_to_finish=False)
            xmp.wait()

        '''wait until all processes are done'''
        xmp.waitForCompletion()

        xmp.refresh()
        err_cnt = dict(n_timeout=0, n_errors=0)

        print('\n___Summary___')
        for index, process in enumerate(xmp.processes):
            status, msg, runTime = xmp.analyzeProcessStatus(process, err_cnt)

            self.junit_build.addTestCase(disp_paths[index], runTime, msg)

            print('Test {}/{} ==> {} ==> {}'.format(index + 1, n_tests, disp_paths[index], status))

        print('Total Timeout: {}/{}'.format(err_cnt['n_timeout'], n_tests))
        print('Total Errors: {}/{}'.format(err_cnt['n_errors'], n_tests))
        print('Total Failed: {}/{}'.format(err_cnt['n_timeout'] + err_cnt['n_errors'], n_tests))

        '''save at the end of the loop'''
        self.junit_build.writeXML()


if __name__ == '__main__':
    '''Quickstart for Debugging / Developing'''
    sp = os.path.join(Paths.getComponentSpaceRoot(), 'Simulation', 'Motor', 'Magnetic', 'tests', 'PMSM')
    os.chdir(sp)
    test_builder = ProjectTestBuilder(sp)
    # test_builder.startBuild()
    test_builder.deployReports()
