#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 3465 $
$LastChangedDate:: 2020-08-25 09:49:49 +0200#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

from collections import deque


def kahn_topsort(graph: dict) -> list:
    """
    Topological sort of dependencies with Kahn's algorithm

    Parameters
    ----------
    graph : dict
        key: node
        value: list of dependent nodes

    Returns
    -------
    ordered_nodes : list
        list of ordered nodes
    """
    in_degree = {u: 0 for u in graph}
    for u in graph:
        '''determine in-degree of each node'''
        for v in graph[u]:
            in_degree[v] += 1

    Q = deque()
    for u in in_degree:
        if in_degree[u] == 0:
            '''collect nodes with zero in-degree'''
            Q.appendleft(u)

    '''list for order of nodes'''
    ordered_nodes = []

    while Q:
        '''choose node of zero in-degree and 'remove' it from graph'''
        u = Q.pop()
        ordered_nodes.append(u)
        for v in graph[u]:
            in_degree[v] -= 1
            if in_degree[v] == 0:
                Q.appendleft(v)

    if len(ordered_nodes) == len(graph):
        ordered_nodes.reverse()
        return ordered_nodes
    else:
        raise Exception('Cycle detected!')
        # '''if there is a cycle, then return an empty list'''
        # return []


def dfs_topsort(graph: dict) -> list:
    """
    Topological sort of dependencies with recursive Depth-first algorithm

    Parameters
    ----------
    graph : dict
        key: node
        value: list of dependent nodes

    Returns
    -------
    ordered_nodes : list
        list of ordered nodes
    """
    '''list for order of nodes'''
    ordered_nodes = []
    color = {u: "white" for u in graph}
    found_cycle = [False]
    for u in graph:
        if color[u] == "white":
            dfs_visit(graph, u, color, ordered_nodes, found_cycle)
        if found_cycle[0]:
            break

    if found_cycle[0]:
        # '''if there is a cycle, then return an empty list'''
        # ordered_nodes = []
        raise Exception('Cycle detected!')

    return ordered_nodes


def dfs_visit(graph, u, color, ordered_nodes, found_cycle):
    if found_cycle[0]:
        return
    color[u] = "gray"
    for v in graph[u]:
        if color[v] == "gray":
            found_cycle[0] = True
            return
        if color[v] == "white":
            dfs_visit(graph, v, color, ordered_nodes, found_cycle)
    color[u] = "black"
    ordered_nodes.append(u)


if __name__ == '__main__':
    graph_tasks = {"Stator Interior1": ['Steelstacking', 'WindingTurnConfig1', 'MagnetWireRound', 'Winding'],
                   "Stator Interior2": ['Steelstacking', 'WindingTurnConfig2', 'MagnetWireRound', 'Winding'],
                   "Steelstacking": [],
                   "WindingTurnConfig1": [],
                   "WindingTurnConfig2": [],
                   "MagnetWireRound": [],
                   "Winding": [],
                   "MagnetGap": [],
                   "Rotor Interior1": ['Steelstacking', 'MagnetGap'],
                   "Rotor Interior2": ['Steelstacking', 'MagnetGap']}

    # order = dfs_topsort(graph_tasks)
    order = kahn_topsort(graph_tasks)

    for task in order:
        print(task)
